# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

# electron mass
me = 0.00054858  # u

# Electron charge
qe = 1.60217733e-19  # A*s

# Avogadro Number
NA = 6.0221367e23  # 1/mol

# Boltzmann
kb = 1.38064852e-23  # J/K

# gas constant
R = 8.3144598  # J/K

# factor between the sigma and FWHM for Gaussian
gfact = 2.3548200450309493
